extends CharacterBody2D

signal remove_from_array(object)

@export var movement_speed = 20.0
@export var hp = 10
@export var knockback_recovery = 3.5;
@export var experience = 1
@export var enemy_damage = 1

var knockback = Vector2.ZERO
var death_animation = preload("res://Enemy/explosion.tscn")
var experience_gem = preload("res://Objects/experience_gem.tscn")

@onready var player = get_tree().get_first_node_in_group("player")
@onready var loot_base = get_tree().get_first_node_in_group("loot")
# Done to make it easy to update references if the path changes.
@onready var sprite = $Sprite2D
@onready var animation = $AnimationPlayer
@onready var sound_hit = $HitSound
@onready var hit_box = $HitBox

func _ready():
	animation.play("walk")
	hit_box.damage = enemy_damage


func _physics_process(_delta):
	knockback = knockback.move_toward(Vector2.ZERO, knockback_recovery)
	var direction = global_position.direction_to(player.global_position)
	velocity = direction * movement_speed
	velocity += knockback
	move_and_slide()

	# 0.1 to stop more frequent flipping.
	if direction.x > 0.1:
		sprite.flip_h = true
	elif direction.x < -0.1:
		sprite.flip_h = false


func _on_hurt_box_hurt(damage, angle, knockback_amount):
	hp -= damage
	knockback = angle * knockback_amount
	if hp <= 0:
		death()
	else:
		sound_hit.play()


func death():
	remove_from_array.emit(self)
	var enemy_death = death_animation.instantiate() as Sprite2D
	enemy_death.scale = sprite.scale
	enemy_death.global_position = global_position
	get_parent().call_deferred("add_child", enemy_death)

	var new_gem = experience_gem.instantiate()
	new_gem.global_position = global_position
	new_gem.experience = experience
	loot_base.call_deferred("add_child", new_gem)

	queue_free()
