extends CharacterBody2D

signal player_death

var movement_speed = 40.0
var hp = 80
var max_hp = 80
var last_movement = Vector2.UP
var experience = 0
var experience_level = 1
var collected_experience = 0
var time = 0

# Attacks
var ice_spear = preload("res://Player/Attack/ice_spear.tscn")
var tornado = preload("res://Player/Attack/tornado.tscn")
var javelin = preload("res://Player/Attack/javelin.tscn")

var ice_spear_ammo = 0
var ice_spear_base_ammo = 0
var ice_spear_attack_speed = 1.5
var ice_spear_level = 0

var tornado_ammo = 0
var tornado_base_ammo = 0
var tornado_attack_speed = 3
var tornado_level = 0

var javelin_ammo = 0
var javelin_level = 0

var enemy_close = []

var collected_upgrades = []
var upgrade_choices = []
var armor = 0
var speed = 0
var spell_cooldown = 0
var spell_size = 0
var additional_attacks = 0

@onready var sprite = $Sprite2D
@onready var walk_timer = get_node("%WalkTimer")
@onready var ice_spear_timer: Timer = $Attack/IceSpearTimer
@onready var ice_spear_attack_timer: Timer = $Attack/IceSpearTimer/IceSpearAttackTimer
@onready var tornado_timer: Timer = $Attack/TornadoTimer
@onready var tornado_attack_timer: Timer = $Attack/TornadoTimer/TornadoAttackTimer
@onready var javelin_base = $Attack/JavelinBase

@onready var experience_bar = $GUILayer/GUI/ExperienceBar
@onready var level_label = $GUILayer/GUI/ExperienceBar/LevelLabel
@onready var level_panel = $GUILayer/GUI/LevelUp
@onready var upgrade_options = $GUILayer/GUI/LevelUp/UpgradeOptions
@onready var item_options= preload("res://Utility/item_option.tscn")
@onready var level_up_sound = $GUILayer/GUI/LevelUp/LevelUpSound
@onready var health_bar = $GUILayer/GUI/HealthBar
@onready var timer_label = $GUILayer/GUI/TimerLabel
@onready var gui_collected_weapons = $GUILayer/GUI/CollectedWeapons
@onready var gui_collected_upgrades = $GUILayer/GUI/CollectedUpgrades
@onready var item_container = preload("res://Player/GUI/item_container.tscn")

@onready var death_panel = $GUILayer/GUI/DeathPanel
@onready var result_label = $GUILayer/GUI/DeathPanel/ResultLabel
@onready var victory_sound = $GUILayer/GUI/DeathPanel/VictorySound
@onready var lose_sound = $GUILayer/GUI/DeathPanel/LoseSound


func _ready():
	# Give character an initial attack.
	upgrade_character("icespear1")
	attack()
	set_experience_bar(experience, calculate_experience_cap())
	_on_hurt_box_hurt(0, 0, 0)


func _physics_process(delta):
	movement()


func movement():
	var x_mov = Input.get_action_strength("right") - Input.get_action_strength("left")
	var y_mov = Input.get_action_strength("down") - Input.get_action_strength("up")
	var mov = Vector2(x_mov, y_mov)
	if mov.x > 0:
		sprite.flip_h = true
	elif mov.x < 0:
		sprite.flip_h = false

	if mov != Vector2.ZERO:
		last_movement = mov
		if walk_timer.is_stopped():
			if sprite.frame >= sprite.hframes - 1:
				sprite.frame = 0
			else:
				sprite.frame += 1
			walk_timer.start()

	velocity = mov.normalized() * movement_speed
	move_and_slide()


func attack():
	if ice_spear_level > 0:
		ice_spear_timer.wait_time = ice_spear_attack_speed * (1 - spell_cooldown)
		if ice_spear_timer.is_stopped():
			ice_spear_timer.start()
	if tornado_level > 0:
		tornado_timer.wait_time = tornado_attack_speed * (1 - spell_cooldown)
		if tornado_timer.is_stopped():
			tornado_timer.start()
	if javelin_level > 0:
		spawn_javelin()


func _on_hurt_box_hurt(damage, _angle, _knockback):
	hp -= clamp(damage - armor, 1, 999)
	health_bar.max_value = max_hp
	health_bar.value = hp
	if hp <= 0:
		death()


func _on_ice_spear_timer_timeout():
	# Loads ammo.
	ice_spear_ammo += ice_spear_base_ammo + additional_attacks
	ice_spear_attack_timer.start()


func _on_ice_spear_attack_timer_timeout():
	if ice_spear_ammo > 0:
		var ice_spear_attack = ice_spear.instantiate()
		ice_spear_attack.position = position
		ice_spear_attack.target = get_random_target()
		#ice_spear_attack.target = _get_closest_target()
		ice_spear_attack.level = ice_spear_level
		add_child(ice_spear_attack)
		ice_spear_ammo -= 1
		if ice_spear_ammo > 0:
			ice_spear_attack_timer.start()
		else:
			ice_spear_attack_timer.stop()


func _on_tornado_timer_timeout():
	tornado_ammo += tornado_base_ammo + additional_attacks
	tornado_attack_timer.start()


func _on_tornado_attack_timer_timeout():
	if tornado_ammo > 0:
		var tornado_attack = tornado.instantiate()
		tornado_attack.position = position
		tornado_attack.last_movement = last_movement
		tornado_attack.level = tornado_level
		add_child(tornado_attack)
		tornado_ammo -= 1
		if tornado_ammo > 0:
			tornado_attack_timer.start()
		else:
			tornado_attack_timer.stop()


func get_random_target():
	if enemy_close.size() > 0:
		return enemy_close.pick_random().global_position
	else:
		return Vector2.UP


func _get_closest_target() -> Vector2:
	if enemy_close.size() == 0:
		return Vector2.UP
	var closest_distance = INF
	var closest_enemy
	for enemy in enemy_close:
		var distance = (global_position - enemy.global_position).length()
		if distance < closest_distance:
			closest_distance = distance
			closest_enemy = enemy
	return closest_enemy.global_position

func _on_enemy_detection_area_body_entered(body):
	if not enemy_close.has(body):
		enemy_close.append(body)


func _on_enemy_detection_area_body_exited(body):
	if enemy_close.has(body):
		enemy_close.erase(body)


func spawn_javelin():
	var get_javelin_total = javelin_base.get_child_count()
	var calc_spawns = (javelin_ammo + additional_attacks) - get_javelin_total
	while calc_spawns > 0:
		var javelin_spawn = javelin.instantiate()
		javelin_spawn.global_position = global_position
		javelin_base.add_child(javelin_spawn)
		calc_spawns -= 1
	# Update javelin.
	var get_javelins = javelin_base.get_children()
	for i in get_javelins:
		if i.has_method("update_javelin"):
			i.update_javelin()


func calculate_experience(gem_experience):
	var exp_required = calculate_experience_cap()
	collected_experience += gem_experience
	if experience + collected_experience >= exp_required:
		collected_experience -= exp_required - experience
		experience_level += 1
		experience = 0
		exp_required = calculate_experience_cap()
		level_up()
	else:
		experience += collected_experience
		collected_experience = 0
	set_experience_bar(experience, exp_required)


func calculate_experience_cap():
	var experience_cap = experience_level
	if experience_level < 20:
		experience_cap = experience_level * 5
	elif experience_level < 40:
		experience_cap = 95 * (experience_level - 19) * 8
	else:
		experience_cap = 255 + (experience_level - 39) * 12
	return experience_cap


func set_experience_bar(set_value = 1, set_max_value = 100):
	experience_bar.max_value = set_max_value
	experience_bar.value = set_value


func level_up():
	level_up_sound.play()
	level_label.text = str("Level: ", experience_level)
	var tween = level_panel.create_tween()
	tween.tween_property(level_panel, "position", Vector2(220, 50), 0.2).set_trans(Tween.TRANS_QUINT).set_ease(Tween.EASE_IN)
	level_panel.visible = true
	var options = 0
	var options_max = 3
	while options < options_max:
		var option_choice = item_options.instantiate()
		option_choice.item = get_random_item()
		upgrade_options.add_child(option_choice)
		options += 1
	get_tree().paused = true


func upgrade_character(upgrade):
	match upgrade:
		"icespear1":
			ice_spear_level = 1
			ice_spear_base_ammo += 1
		"icespear2":
			ice_spear_level = 2
			ice_spear_base_ammo += 1
		"icespear3":
			ice_spear_level = 3
		"icespear4":
			ice_spear_level = 4
			ice_spear_base_ammo += 2
		"tornado1":
			tornado_level = 1
			tornado_base_ammo += 1
		"tornado2":
			tornado_level = 2
			tornado_base_ammo += 1
		"tornado3":
			tornado_level = 3
			tornado_attack_speed -= 0.5
		"tornado4":
			tornado_level = 4
			tornado_base_ammo += 1
		"javelin1":
			javelin_level = 1
			javelin_ammo = 1
		"javelin2":
			javelin_level = 2
		"javelin3":
			javelin_level = 3
		"javelin4":
			javelin_level = 4
		"armor1","armor2","armor3","armor4":
			armor += 1
		"speed1","speed2","speed3","speed4":
			movement_speed += 20.0
		"tome1","tome2","tome3","tome4":
			spell_size += 0.10
		"scroll1","scroll2","scroll3","scroll4":
			spell_cooldown += 0.05
		"ring1","ring2":
			additional_attacks += 1
		"food":
			hp += 20
			hp = clamp(hp,0,max_hp)
	adjust_gui_collection(upgrade)
	attack()
	var option_children = upgrade_options.get_children()
	for i in option_children:
		i.queue_free()
	upgrade_choices.clear()
	collected_upgrades.append(upgrade)
	level_panel.visible = false
	level_panel.position = Vector2(800, 50)
	get_tree().paused = false
	calculate_experience(0)


func get_random_item():
	var db_list = []
	for i in UpgradeDb.UPGRADES:
		if i in collected_upgrades:
			pass
		elif i in upgrade_choices:
			pass
		elif UpgradeDb.UPGRADES[i]["type"] == "item":
			pass
		elif UpgradeDb.UPGRADES[i]["prerequisite"].size() > 0:
			var to_add = true
			for n in UpgradeDb.UPGRADES[i]["prerequisite"]:
				if not n in collected_upgrades:
					to_add = false
					pass
			if to_add:
				db_list.append(i)
		else:
			db_list.append(i)
	if db_list.size() > 0:
		var random_item = db_list.pick_random()
		upgrade_choices.append(random_item)
		return random_item
	else:
		return null


func _on_grab_area_area_entered(area):
	if area.is_in_group("loot"):
		area.target = self


func _on_collect_area_area_entered(area):
	if area.is_in_group("loot"):
		var gem_experience = area.collect()
		calculate_experience(gem_experience)

func change_time(arg_time = 0):
	time = arg_time
	var get_minutes = int(time / 60)
	var get_seconds = time % 60
	if get_minutes < 10:
		get_minutes = str(0, get_minutes)
	if get_seconds < 10:
		get_seconds = str(0, get_seconds)
	timer_label.text = str(get_minutes, ":", get_seconds)


func adjust_gui_collection(upgrade):
	var get_upgraded_display_name = UpgradeDb.UPGRADES[upgrade]["displayname"]
	var get_type = UpgradeDb.UPGRADES[upgrade]["type"]
	if get_type != "item":
		var get_collected_display_names = []
		for i in collected_upgrades:
			get_collected_display_names.append(UpgradeDb.UPGRADES[i]["displayname"])
		if not get_upgraded_display_name in get_collected_display_names:
			var new_item = item_container.instantiate()
			new_item.upgrade = upgrade
			match get_type:
				"weapon":
					gui_collected_weapons.add_child(new_item)
				"upgrade":
					gui_collected_upgrades.add_child(new_item)


func death():
	death_panel.visible = true
	player_death.emit()
	get_tree().paused = true
	var tween = death_panel.create_tween()
	tween.tween_property(death_panel, "position", Vector2(220, 50), 3).set_trans(Tween.TRANS_QUINT).set_ease(Tween.EASE_OUT)
	if time >= 300:
		result_label.text = "You Win"
		victory_sound.play()
	else:
		result_label.text = "You Lose"
		lose_sound.play()


func _on_menu_button_click_end():
	get_tree().paused = false
	get_tree().change_scene_to_file("res://TitleScreen/menu.tscn")
